/*
 *	Copyright (c) 2024 Scott Duensing, scott@kangaroopunch.com
 *
 *	Permission is hereby granted, free of charge, to any person obtaining a copy
 *	of this software and associated documentation files (the "Software"), to deal
 *	in the Software without restriction, including without limitation the rights
 *	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *	copies of the Software, and to permit persons to whom the Software is
 *	furnished to do so, subject to the following conditions:
 *
 *	The above copyright notice and this permission notice shall be included in
 *	all copies or substantial portions of the Software.
 *
 *	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 *	SOFTWARE.
 */


#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>

#include "portme.h"
#include "story.h"
#include "state.h"
#include "interpreter.h"
#include "text.h"

#pragma push_macro("bool")
#undef bool
#define DOS_IMPLEMENTATION
#include "dos.h"
#pragma pop_macro("bool")


static char    *_storyFile = NULL;
static uint8_t *_RAM       = NULL;
static uint8_t _attr       = 7;


void portAttributeSet(byte attribute) {
	switch (attribute) {
		case TEXT_ATTR_NORMAL:
			textbackground(1);
			textcolor(7);
			_attr = (1 << 4) + 7;
			break;

		case TEXT_ATTR_REVERSE:
			textbackground(7);
			textcolor(1);
			_attr = (7 << 4) + 1;
			break;

		case TEXT_ATTR_BOLD:
			textbackground(1);
			textcolor(15);
			_attr = (1 << 4) + 15;
			break;

		case TEXT_ATTR_EMPHASIS:
			textbackground(1);
			textcolor(14);
			_attr = (1 << 4) + 14;
			break;
	}
}


uint8_t portByteGet(uint32_t address) {
	return _RAM[address];
}


void portByteSet(uint32_t address, uint8_t value) {
	_RAM[address] = value;
}


char portCharGet(void) {
	char           c;
	enum keycode_t key;

#if 0
	static char     playback[] = "open mailbox\ntake leaflet\nread leaflet\ndrop leaflet\n";
	static uint32_t pointer = 0;

	if (pointer < strlen(playback)) return playback[pointer++];
#endif

	do {
		c   = *readchars();
		key = *readkeys();
		if (key == KEY_RETURN) c = 13;
		if (key == KEY_BACK)   c = 127;
		if (key == KEY_DELETE) c = 8;
	} while (c == 0);

	return c;
}


void portCharGetPos(byte *x, byte *y) {
	*x = wherex() + 1;
	*y = wherey() + 1;
}


void portCharPrint(char c) {
	char      ch[2] = { 0, 0 };
	uint16_t *p;
	uint16_t  v = (_attr << 8) + ' ';
	byte      x;

	if ((wherex() >= screenwidth()) || (c == '\n')) {
		gotoxy(0, wherey() + 1);
		while (wherey() >= screenheight()) {
			memmove(((uint16_t*)screenbuffer()), ((uint16_t*)screenbuffer()) + screenwidth(), screenwidth() * (screenheight() - 1) * sizeof(uint16_t));
			p = (uint16_t *)screenbuffer() + (screenwidth() * (screenheight() - 1));
			for (x=0; x<screenwidth(); x++) *p++ = v;
			gotoxy(0, wherey() - 1);
		}
		return;
	}

	ch[0] = c;
	cputs(ch);
}


void portCharSetPos(byte x, byte y) {
	gotoxy(x - 1, y - 1);
}


void portDie(char *fmt, ...) {
	va_list ap;

	printf("\n");
	va_start(ap, fmt);
	vprintf(fmt, ap);
	va_end(ap);
	printf("\n");

	exit(1);
}


bool portFileRestore(void) {

	FILE *in;
	bool  ok  = false;

	in = fopen("save.dat", "rb");
	if (in) {
		ok = true;

		// Read in PC.
		ok &= (fread(&__state.pc, sizeof(__state.pc), 1, in) == 1);
		// Read in SP.
		ok &= (fread(&__state.sp, sizeof(__state.sp), 1, in) == 1);
		// Read in BP.
		ok &= (fread(&__state.bp, sizeof(__state.bp), 1, in) == 1);
		// Read in dynamic game RAM.
		ok &= (fread(_RAM, storyStaticMemoryBaseAddress(), 1, in) == 1);
		// Read in stack.
		ok &= (fread(__state.stack, sizeof(__state.stack[0]), __state.sp, in) == __state.sp);

		fclose(in);
	}

	return ok;
}


bool portFileSave(void) {
	FILE *out;
	bool  ok  = false;

	out = fopen("save.dat", "wb");
	if (out) {
		ok = true;

		// Write out PC.
		ok &= (fwrite(&__state.pc, sizeof(__state.pc), 1, out) == 1);
		// Write out SP.
		ok &= (fwrite(&__state.sp, sizeof(__state.sp), 1, out) == 1);
		// Write out BP.
		ok &= (fwrite(&__state.bp, sizeof(__state.bp), 1, out) == 1);
		// Write out dynamic game RAM.
		ok &= (fwrite(_RAM, storyStaticMemoryBaseAddress(), 1, out) == 1);
		// Write out stack.
		ok &= (fwrite(__state.stack, sizeof(__state.stack[0]), __state.sp, out) == __state.sp);

		fclose(out);
	}

	return ok;
}


uint16_t portRandomGet(int16_t range) {
	// If range is zero, randomize with "best" random seed.
	// If range is negative, use the positive value to seed.

	if (range <= 0) {
		srandom(range < 0 ? -range : time(NULL));
		return 0;
	}

	return (random() % range) + 1;
}


void portStoryLoad(void) {
	FILE     *in;
	uint32_t  length;

	in = fopen(_storyFile, "rb");
	if (!in) portDie("Unable to open %s!\n", _storyFile);

	fseek(in, 0, SEEK_END);
	length = ftell(in);
	fseek(in, 0, SEEK_SET);

	_RAM = (byte *)malloc(length);
	if (!_RAM) {
		fclose(in);
		portDie("Unable to allocate %u bytes!\n", length);
	}

	if (fread(_RAM, length, 1, in) != 1) {
		free(_RAM);
		fclose(in);
		portDie("Unable to read %s!\n", _storyFile);
	}

	fclose(in);

	storyChecksumCalculate();
}


uint16_t portWordGet(uint32_t address) {
	return ((uint16_t)_RAM[address] << 8) | ((uint16_t)_RAM[address + 1]);
}


void portWordSet(uint32_t address, uint16_t value) {
	_RAM[address]     = (value >> 8) & 0xff;   // MSB first.
	_RAM[address + 1] = value & 0xff;
}


int main(int argc, char *argv[]) {

	if (argc != 2) portDie("Usage:  %s [storyfile]\n", argv[0]);

	_storyFile = argv[1];

	setvideomode(videomode_80x25_9x16);
	curson();
	textbackground(1);
	textcolor(7);
	clrscr();
	gotoxy(0, screenheight() - 1);

	stateReset();
	portStoryLoad();
	opcodesSetup();
	storySetup(80, 25);
	interpreterRun();

	free(_RAM);

	return 0;
}
